from typing import Optional, Sequence

import numpy as np


class GraphData:
    """GraphData class

    This data class is almost same as `torch_geometric.data.Data
    <https://pytorch-geometric.readthedocs.io/en/latest/modules/data.html#torch_geometric.data.Data>`_.

    Attributes
    ----------
    node_features: np.ndarray
        Node feature matrix with shape [num_nodes, num_node_features]
    edge_index: np.ndarray, dtype int
        Graph connectivity in COO format with shape [2, num_edges]
    edge_features: np.ndarray, optional (default None)
        Edge feature matrix with shape [num_edges, num_edge_features]
    node_pos_features: np.ndarray, optional (default None)
        Node position matrix with shape [num_nodes, num_dimensions].
    num_nodes: int
        The number of nodes in the graph
    num_node_features: int
        The number of features per node in the graph
    num_edges: int
        The number of edges in the graph
    num_edges_features: int, optional (default None)
        The number of features per edge in the graph

    Examples
    --------
    >>> import numpy as np
    >>> node_features = np.random.rand(5, 10)
    >>> edge_index = np.array([[0, 1, 2, 3, 4], [1, 2, 3, 4, 0]], dtype=np.int64)
    >>> edge_features = np.random.rand(5, 5)
    >>> global_features = np.random.random(5)
    >>> graph = GraphData(node_features, edge_index, edge_features, z=global_features)
    >>> graph
    GraphData(node_features=[5, 10], edge_index=[2, 5], edge_features=[5, 5], z=[5])
    """

    def __init__(self,
                 node_features: np.ndarray,
                 edge_index: np.ndarray,
                 edge_features: Optional[np.ndarray] = None,
                 node_pos_features: Optional[np.ndarray] = None,
                 **kwargs):
        """
        Parameters
        ----------
        node_features: np.ndarray
            Node feature matrix with shape [num_nodes, num_node_features]
        edge_index: np.ndarray, dtype int
            Graph connectivity in COO format with shape [2, num_edges]
        edge_features: np.ndarray, optional (default None)
            Edge feature matrix with shape [num_edges, num_edge_features]
        node_pos_features: np.ndarray, optional (default None)
            Node position matrix with shape [num_nodes, num_dimensions].
        kwargs: optional
            Additional attributes and their values
        """
        # validate params
        if isinstance(node_features, np.ndarray) is False:
            raise ValueError('node_features must be np.ndarray.')

        if isinstance(edge_index, np.ndarray) is False:
            raise ValueError('edge_index must be np.ndarray.')
        elif issubclass(edge_index.dtype.type, np.integer) is False:
            raise ValueError('edge_index.dtype must contains integers.')
        elif edge_index.shape[0] != 2:
            raise ValueError('The shape of edge_index is [2, num_edges].')

        # np.max() method works only for a non-empty array, so size of the array should be non-zero
        elif (edge_index.size != 0) and (np.max(edge_index) >=
                                         len(node_features)):
            raise ValueError('edge_index contains the invalid node number.')

        if edge_features is not None:
            if isinstance(edge_features, np.ndarray) is False:
                raise ValueError('edge_features must be np.ndarray or None.')
            elif edge_index.shape[1] != edge_features.shape[0]:
                raise ValueError(
                    'The first dimension of edge_features must be the \
                          same as the second dimension of edge_index.')

        if node_pos_features is not None:
            if isinstance(node_pos_features, np.ndarray) is False:
                raise ValueError(
                    'node_pos_features must be np.ndarray or None.')
            elif node_pos_features.shape[0] != node_features.shape[0]:
                raise ValueError(
                    'The length of node_pos_features must be the same as the \
                          length of node_features.')

        self.node_features = node_features
        self.edge_index = edge_index
        self.edge_features = edge_features
        self.node_pos_features = node_pos_features
        self.kwargs = kwargs
        self.num_nodes, self.num_node_features = self.node_features.shape
        self.num_edges = edge_index.shape[1]
        if self.edge_features is not None:
            self.num_edge_features = self.edge_features.shape[1]

        for key, value in self.kwargs.items():
            setattr(self, key, value)

    def __repr__(self) -> str:
        """Returns a string containing the printable representation of the object"""
        cls = self.__class__.__name__
        node_features_str = str(list(self.node_features.shape))
        edge_index_str = str(list(self.edge_index.shape))
        if self.edge_features is not None:
            edge_features_str = str(list(self.edge_features.shape))
        else:
            edge_features_str = "None"

        out = "%s(node_features=%s, edge_index=%s, edge_features=%s" % (
            cls, node_features_str, edge_index_str, edge_features_str)
        # Adding shapes of kwargs
        for key, value in self.kwargs.items():
            if isinstance(value, np.ndarray):
                out += (', ' + key + '=' + str(list(value.shape)))
            elif isinstance(value, str):
                out += (', ' + key + '=' + value)
            elif isinstance(value, int) or isinstance(value, float):
                out += (', ' + key + '=' + str(value))
        out += ')'
        return out

    def to_pyg_graph(self):
        """Convert to PyTorch Geometric graph data instance

        Returns
        -------
        torch_geometric.data.Data
            Graph data for PyTorch Geometric

        Note
        ----
        This method requires PyTorch Geometric to be installed.
        """
        try:
            import torch
            from torch_geometric.data import Data
        except ModuleNotFoundError:
            raise ImportError(
                "This function requires PyTorch Geometric to be installed.")

        edge_features = self.edge_features
        if edge_features is not None:
            edge_features = torch.from_numpy(self.edge_features).float()
        node_pos_features = self.node_pos_features
        if node_pos_features is not None:
            node_pos_features = torch.from_numpy(self.node_pos_features).float()
        kwargs = {}
        for key, value in self.kwargs.items():
            kwargs[key] = torch.from_numpy(value).float()
        return Data(x=torch.from_numpy(self.node_features).float(),
                    edge_index=torch.from_numpy(self.edge_index).long(),
                    edge_attr=edge_features,
                    pos=node_pos_features,
                    **kwargs)

    def to_dgl_graph(self, self_loop: bool = False):
        """Convert to DGL graph data instance

        Returns
        -------
        dgl.DGLGraph
            Graph data for DGL
        self_loop: bool
            Whether to add self loops for the nodes, i.e. edges from nodes
            to themselves. Default to False.

        Note
        ----
        This method requires DGL to be installed.
        """
        try:
            import dgl
            import torch
        except ModuleNotFoundError:
            raise ImportError("This function requires DGL to be installed.")

        src = self.edge_index[0]
        dst = self.edge_index[1]

        g = dgl.graph(
            (torch.from_numpy(src).long(), torch.from_numpy(dst).long()),
            num_nodes=self.num_nodes)
        g.ndata['x'] = torch.from_numpy(self.node_features).float()

        if self.node_pos_features is not None:
            g.ndata['pos'] = torch.from_numpy(self.node_pos_features).float()

        if self.edge_features is not None:
            g.edata['edge_attr'] = torch.from_numpy(self.edge_features).float()

        if self_loop:
            # This assumes that the edge features for self loops are full-zero tensors
            # In the future we may want to support featurization for self loops
            g.add_edges(np.arange(self.num_nodes), np.arange(self.num_nodes))

        return g

    def numpy_to_torch(self, device: str = 'cpu'):
        """Convert numpy arrays to torch tensors. This may be useful when you are using PyTorch Geometric with GraphData objects.

        Parameters
        ----------
        device : str
            Device to store the tensors. Default to 'cpu'.

        Example
        -------
        >>> num_nodes, num_node_features = 5, 32
        >>> num_edges, num_edge_features = 6, 32
        >>> node_features = np.random.random_sample((num_nodes, num_node_features))
        >>> edge_features = np.random.random_sample((num_edges, num_edge_features))
        >>> edge_index = np.random.randint(0, num_nodes, (2, num_edges))
        >>> graph_data = GraphData(node_features, edge_index, edge_features)
        >>> graph_data = graph_data.numpy_to_torch()
        >>> print(type(graph_data.node_features))
        <class 'torch.Tensor'>
        """
        import copy

        import torch
        graph_copy = copy.deepcopy(self)

        graph_copy.node_features = torch.from_numpy(  # type: ignore
            self.node_features).float().to(device)
        graph_copy.edge_index = torch.from_numpy(  # type: ignore
            self.edge_index).long().to(device)
        if self.edge_features is not None:
            graph_copy.edge_features = torch.from_numpy(  # type: ignore
                self.edge_features).float().to(device)
        else:
            graph_copy.edge_features = None
        if self.node_pos_features is not None:
            graph_copy.node_pos_features = torch.from_numpy(  # type: ignore
                self.node_pos_features).float().to(device)
        else:
            graph_copy.node_pos_features = None

        graph_copy.kwargs = {}
        for key, value in self.kwargs.items():
            if isinstance(value, np.ndarray):
                value = torch.from_numpy(value).to(device)
                graph_copy.kwargs[key] = value
                setattr(graph_copy, key, value)

        return graph_copy


class BatchGraphData(GraphData):
    """Batch GraphData class

    Attributes
    ----------
    node_features: np.ndarray
        Concatenated node feature matrix with shape [num_nodes, num_node_features].
        `num_nodes` is total number of nodes in the batch graph.
    edge_index: np.ndarray, dtype int
        Concatenated graph connectivity in COO format with shape [2, num_edges].
        `num_edges` is total number of edges in the batch graph.
    edge_features: np.ndarray, optional (default None)
        Concatenated edge feature matrix with shape [num_edges, num_edge_features].
        `num_edges` is total number of edges in the batch graph.
    node_pos_features: np.ndarray, optional (default None)
        Concatenated node position matrix with shape [num_nodes, num_dimensions].
        `num_nodes` is total number of edges in the batch graph.
    num_nodes: int
        The number of nodes in the batch graph.
    num_node_features: int
        The number of features per node in the graph.
    num_edges: int
        The number of edges in the batch graph.
    num_edges_features: int, optional (default None)
        The number of features per edge in the graph.
    graph_index: np.ndarray, dtype int
        This vector indicates which graph the node belongs with shape [num_nodes,].

    Examples
    --------
    >>> import numpy as np
    >>> from deepchem.feat.graph_data import GraphData
    >>> node_features_list = np.random.rand(2, 5, 10)
    >>> edge_index_list = np.array([
    ...    [[0, 1, 2, 3, 4], [1, 2, 3, 4, 0]],
    ...    [[0, 1, 2, 3, 4], [1, 2, 3, 4, 0]],
    ... ], dtype=int)
    >>> user_defined_attribute = np.array([0, 1])
    >>> graph_list = [GraphData(node_features, edge_index, attribute=user_defined_attribute)
    ...     for node_features, edge_index in zip(node_features_list, edge_index_list)]
    >>> batch_graph = BatchGraphData(graph_list=graph_list)
    """

    def __init__(self, graph_list: Sequence[GraphData]):
        """
        Parameters
        ----------
        graph_list: Sequence[GraphData]
            List of GraphData
        """
        # stack features
        batch_node_features = np.vstack(
            [graph.node_features for graph in graph_list])

        # before stacking edge_features or node_pos_features,
        # we should check whether these are None or not
        if graph_list[0].edge_features is not None:
            batch_edge_features: Optional[np.ndarray] = np.vstack(
                [graph.edge_features for graph in graph_list])  # type: ignore
        else:
            batch_edge_features = None

        if graph_list[0].node_pos_features is not None:
            batch_node_pos_features: Optional[np.ndarray] = np.vstack([
                graph.node_pos_features for graph in graph_list  # type: ignore
            ])
        else:
            batch_node_pos_features = None

        # create new edge index
        # number of nodes in each graph
        num_nodes_list = [graph.num_nodes for graph in graph_list]
        # cumulative number of nodes for each graph. This is necessary because the values in edge_index are node indices of all of the graphs in graph_list and so we need to offset the indices by the number of nodes in the previous graphs.
        cum_num_nodes_list = np.cumsum([0] + num_nodes_list)[:-1]
        # columns are the edge index, values are the node index
        batch_edge_index = np.hstack([
            graph.edge_index + cum_num_nodes
            for cum_num_nodes, graph in zip(cum_num_nodes_list, graph_list)
        ])

        # graph_index indicates which nodes belong to which graph
        graph_index = []
        for i, num_nodes in enumerate(num_nodes_list):
            graph_index.extend([i] * num_nodes)
        self.graph_index = np.array(graph_index)

        # Batch user defined attributes
        kwargs = {}
        user_defined_attribute_names = self._get_user_defined_attributes(
            graph_list[0])
        for name in user_defined_attribute_names:
            kwargs[name] = np.vstack(
                [getattr(graph, name) for graph in graph_list])

        super().__init__(node_features=batch_node_features,
                         edge_index=batch_edge_index,
                         edge_features=batch_edge_features,
                         node_pos_features=batch_node_pos_features,
                         **kwargs)

    def _get_user_defined_attributes(self, graph_data: GraphData):
        """A GraphData object can have user defined attributes but the attribute name of those
        are unknown since it can be arbitary. This method helps to find user defined attribute's
        name by making a list of known graph data attributes and finding other user defined
        attributes via `vars` method. The user defined attributes are attributes other than
        `node_features`, `edge_index`, `edge_features`, `node_pos_features`, `kwargs`, `num_nodes`,
        `num_node_features`, `num_edges`, `num_edge_features` as these are graph data attributes."""
        graph_data_attributes = [
            'node_features',
            'edge_index',
            'edge_features',
            'node_pos_features',
            'kwargs',
            'num_nodes',
            'num_node_features',
            'num_edges',
            'num_edge_features',
        ]
        user_defined_attribute_names = []
        for arg in vars(graph_data):
            if arg not in graph_data_attributes:
                user_defined_attribute_names.append(arg)
        return user_defined_attribute_names

    def numpy_to_torch(self, device: str = "cpu"):
        """
        Convert numpy arrays to torch tensors for BatchGraphData. BatchGraphData is very similar to GraphData, but it combines all graphs into a single graph object and it has an additional attribute `graph_index` which indicates which nodes belong to which graph.

        Parameters
        ----------
        device : str
            Device to store the tensors. Default to 'cpu'.

        Example
        -------
        >>> num_nodes, num_node_features = 5, 32
        >>> num_edges, num_edge_features = 6, 32
        >>> node_features = np.random.random_sample((num_nodes, num_node_features))
        >>> edge_features = np.random.random_sample((num_edges, num_edge_features))
        >>> edge_index = np.random.randint(0, num_nodes, (2, num_edges))
        >>> graph_data = GraphData(node_features, edge_index, edge_features)
        >>> node_features2 = np.random.random_sample((num_nodes, num_node_features))
        >>> edge_features2 = np.random.random_sample((num_edges, num_edge_features))
        >>> edge_index2 = np.random.randint(0, num_nodes, (2, num_edges))
        >>> graph_data2 = GraphData(node_features2, edge_index2, edge_features2)
        >>> batch_graph_data = BatchGraphData([graph_data, graph_data2])
        >>> batch_graph_data = batch_graph_data.numpy_to_torch()
        >>> print(type(batch_graph_data.node_features))
        <class 'torch.Tensor'>
        """
        import torch
        graph_copy = super().numpy_to_torch(device)

        graph_index = torch.from_numpy(
            self.graph_index).long().to(  # type: ignore
                device)
        graph_copy.graph_index = graph_index

        return graph_copy



"""
DGL-based GCN for graph property prediction.
"""
import torch.nn as nn
import torch.nn.functional as F

from deepchem.models.losses import Loss, L2Loss, SparseSoftmaxCrossEntropy
from deepchem.models.torch_models.torch_model import TorchModel
from typing import Optional


class GCN(nn.Module):
    """Model for Graph Property Prediction Based on Graph Convolution Networks (GCN).

    This model proceeds as follows:

    * Update node representations in graphs with a variant of GCN
    * For each graph, compute its representation by 1) a weighted sum of the node
        representations in the graph, where the weights are computed by applying a
        gating function to the node representations 2) a max pooling of the node
        representations 3) concatenating the output of 1) and 2)
    * Perform the final prediction using an MLP

    Examples
    --------

    >>> import deepchem as dc
    >>> import dgl
    >>> from deepchem.models import GCN
    >>> smiles = ["C1CCC1", "C1=CC=CN=C1"]
    >>> featurizer = dc.feat.MolGraphConvFeaturizer()
    >>> graphs = featurizer.featurize(smiles)
    >>> print(type(graphs[0]))
    <class 'deepchem.feat.graph_data.GraphData'>
    >>> dgl_graphs = [graphs[i].to_dgl_graph(self_loop=True) for i in range(len(graphs))]
    >>> # Batch two graphs into a graph of two connected components
    >>> batch_dgl_graph = dgl.batch(dgl_graphs)
    >>> model = GCN(n_tasks=1, mode='regression')
    >>> preds = model(batch_dgl_graph)
    >>> print(type(preds))
    <class 'torch.Tensor'>
    >>> preds.shape == (2, 1)
    True

    References
    ----------
    .. [1] Thomas N. Kipf and Max Welling. "Semi-Supervised Classification with Graph
        Convolutional Networks." ICLR 2017.

    Notes
    -----
    This class requires DGL (https://github.com/dmlc/dgl) and DGL-LifeSci
    (https://github.com/awslabs/dgl-lifesci) to be installed.

    This model is different from deepchem.models.GraphConvModel as follows:

    * For each graph convolution, the learnable weight in this model is shared across all nodes.
        ``GraphConvModel`` employs separate learnable weights for nodes of different degrees. A
        learnable weight is shared across all nodes of a particular degree.
    * For ``GraphConvModel``, there is an additional GraphPool operation after each
        graph convolution. The operation updates the representation of a node by applying an
        element-wise maximum over the representations of its neighbors and itself.
    * For computing graph-level representations, this model computes a weighted sum and an
        element-wise maximum of the representations of all nodes in a graph and concatenates them.
        The node weights are obtained by using a linear/dense layer followd by a sigmoid function.
        For ``GraphConvModel``, the sum over node representations is unweighted.
    * There are various minor differences in using dropout, skip connection and batch
        normalization.
    """

    def __init__(self,
                 n_tasks: int,
                 graph_conv_layers: Optional[list] = None,
                 activation=None,
                 residual: bool = True,
                 batchnorm: bool = False,
                 dropout: float = 0.,
                 predictor_hidden_feats: int = 128,
                 predictor_dropout: float = 0.,
                 mode: str = 'regression',
                 number_atom_features: int = 30,
                 n_classes: int = 2,
                 nfeat_name: str = 'x'):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks.
        graph_conv_layers: list of int
            Width of channels for GCN layers. graph_conv_layers[i] gives the width of channel
            for the i-th GCN layer. If not specified, the default value will be [64, 64].
        activation: callable
            The activation function to apply to the output of each GCN layer.
            By default, no activation function will be applied.
        residual: bool
            Whether to add a residual connection within each GCN layer. Default to True.
        batchnorm: bool
            Whether to apply batch normalization to the output of each GCN layer.
            Default to False.
        dropout: float
            The dropout probability for the output of each GCN layer. Default to 0.
        predictor_hidden_feats: int
            The size for hidden representations in the output MLP predictor. Default to 128.
        predictor_dropout: float
            The dropout probability in the output MLP predictor. Default to 0.
        mode: str
            The model type, 'classification' or 'regression'. Default to 'regression'.
        number_atom_features: int
            The length of the initial atom feature vectors. Default to 30.
        n_classes: int
            The number of classes to predict per task
            (only used when ``mode`` is 'classification'). Default to 2.
        nfeat_name: str
            For an input graph ``g``, the model assumes that it stores node features in
            ``g.ndata[nfeat_name]`` and will retrieve input node features from that.
            Default to 'x'.
        """
        try:
            import dgl  # noqa: F401
        except:
            raise ImportError('This class requires dgl.')
        try:
            import dgllife  # noqa: F401
        except:
            raise ImportError('This class requires dgllife.')

        if mode not in ['classification', 'regression']:
            raise ValueError(
                "mode must be either 'classification' or 'regression'")

        super(GCN, self).__init__()

        self.n_tasks = n_tasks
        self.mode = mode
        self.n_classes = n_classes
        self.nfeat_name = nfeat_name
        if mode == 'classification':
            out_size = n_tasks * n_classes
        else:
            out_size = n_tasks

        from dgllife.model import GCNPredictor as DGLGCNPredictor

        if graph_conv_layers is None:
            graph_conv_layers = [64, 64]
        num_gnn_layers = len(graph_conv_layers)

        if activation is not None:
            activation = [activation] * num_gnn_layers

        self.model = DGLGCNPredictor(
            in_feats=number_atom_features,
            hidden_feats=graph_conv_layers,
            activation=activation,
            residual=[residual] * num_gnn_layers,
            batchnorm=[batchnorm] * num_gnn_layers,
            dropout=[dropout] * num_gnn_layers,
            n_tasks=out_size,
            predictor_hidden_feats=predictor_hidden_feats,
            predictor_dropout=predictor_dropout)

    def forward(self, g):
        """Predict graph labels

        Parameters
        ----------
        g: DGLGraph
            A DGLGraph for a batch of graphs. It stores the node features in
            ``dgl_graph.ndata[self.nfeat_name]``.

        Returns
        -------
        torch.Tensor
            The model output.

        * When self.mode = 'regression',
            its shape will be ``(dgl_graph.batch_size, self.n_tasks)``.
        * When self.mode = 'classification', the output consists of probabilities
            for classes. Its shape will be ``(dgl_graph.batch_size, self.n_tasks, self.n_classes)``
            if self.n_tasks > 1; its shape will be ``(dgl_graph.batch_size, self.n_classes)`` if
            self.n_tasks is 1.
        torch.Tensor, optional
            This is only returned when self.mode = 'classification', the output consists of the
            logits for classes before softmax.
        """
        node_feats = g.ndata[self.nfeat_name]
        out = self.model(g, node_feats)

        if self.mode == 'classification':
            if self.n_tasks == 1:
                logits = out.view(-1, self.n_classes)
                softmax_dim = 1
            else:
                logits = out.view(-1, self.n_tasks, self.n_classes)
                softmax_dim = 2
            proba = F.softmax(logits, dim=softmax_dim)
            return proba, logits
        else:
            return out


class GCNModel(TorchModel):
    """Model for Graph Property Prediction Based on Graph Convolution Networks (GCN).

    This model proceeds as follows:

    * Update node representations in graphs with a variant of GCN
    * For each graph, compute its representation by 1) a weighted sum of the node
        representations in the graph, where the weights are computed by applying a
        gating function to the node representations 2) a max pooling of the node
        representations 3) concatenating the output of 1) and 2)
    * Perform the final prediction using an MLP

    Examples
    --------
    >>> import deepchem as dc
    >>> from deepchem.models import GCNModel
    >>> # preparing dataset
    >>> smiles = ["C1CCC1", "CCC"]
    >>> labels = [0., 1.]
    >>> featurizer = dc.feat.MolGraphConvFeaturizer()
    >>> X = featurizer.featurize(smiles)
    >>> dataset = dc.data.NumpyDataset(X=X, y=labels)
    >>> # training model
    >>> model = GCNModel(mode='classification', n_tasks=1,
    ...                  batch_size=16, learning_rate=0.001)
    >>> loss = model.fit(dataset, nb_epoch=5)

    References
    ----------
    .. [1] Thomas N. Kipf and Max Welling. "Semi-Supervised Classification with Graph
        Convolutional Networks." ICLR 2017.

    Notes
    -----
    This class requires DGL (https://github.com/dmlc/dgl) and DGL-LifeSci
    (https://github.com/awslabs/dgl-lifesci) to be installed.

    This model is different from deepchem.models.GraphConvModel as follows:

    * For each graph convolution, the learnable weight in this model is shared across all nodes.
        ``GraphConvModel`` employs separate learnable weights for nodes of different degrees. A
        learnable weight is shared across all nodes of a particular degree.
    * For ``GraphConvModel``, there is an additional GraphPool operation after each
        graph convolution. The operation updates the representation of a node by applying an
        element-wise maximum over the representations of its neighbors and itself.
    * For computing graph-level representations, this model computes a weighted sum and an
        element-wise maximum of the representations of all nodes in a graph and concatenates them.
        The node weights are obtained by using a linear/dense layer followd by a sigmoid function.
        For ``GraphConvModel``, the sum over node representations is unweighted.
    * There are various minor differences in using dropout, skip connection and batch
        normalization.
    """

    def __init__(self,
                 n_tasks: int,
                 graph_conv_layers: Optional[list] = None,
                 activation=None,
                 residual: bool = True,
                 batchnorm: bool = False,
                 dropout: float = 0.,
                 predictor_hidden_feats: int = 128,
                 predictor_dropout: float = 0.,
                 mode: str = 'regression',
                 number_atom_features=30,
                 n_classes: int = 2,
                 self_loop: bool = True,
                 **kwargs):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks.
        graph_conv_layers: list of int
            Width of channels for GCN layers. graph_conv_layers[i] gives the width of channel
            for the i-th GCN layer. If not specified, the default value will be [64, 64].
        activation: callable
            The activation function to apply to the output of each GCN layer.
            By default, no activation function will be applied.
        residual: bool
            Whether to add a residual connection within each GCN layer. Default to True.
        batchnorm: bool
            Whether to apply batch normalization to the output of each GCN layer.
            Default to False.
        dropout: float
            The dropout probability for the output of each GCN layer. Default to 0.
        predictor_hidden_feats: int
            The size for hidden representations in the output MLP predictor. Default to 128.
        predictor_dropout: float
            The dropout probability in the output MLP predictor. Default to 0.
        mode: str
            The model type, 'classification' or 'regression'. Default to 'regression'.
        number_atom_features: int
            The length of the initial atom feature vectors. Default to 30.
        n_classes: int
            The number of classes to predict per task
            (only used when ``mode`` is 'classification'). Default to 2.
        self_loop: bool
            Whether to add self loops for the nodes, i.e. edges from nodes to themselves.
            When input graphs have isolated nodes, self loops allow preserving the original feature
            of them in message passing. Default to True.
        kwargs
            This can include any keyword argument of TorchModel.
        """
        model = GCN(n_tasks=n_tasks,
                    graph_conv_layers=graph_conv_layers,
                    activation=activation,
                    residual=residual,
                    batchnorm=batchnorm,
                    dropout=dropout,
                    predictor_hidden_feats=predictor_hidden_feats,
                    predictor_dropout=predictor_dropout,
                    mode=mode,
                    number_atom_features=number_atom_features,
                    n_classes=n_classes)
        if mode == 'regression':
            loss: Loss = L2Loss()
            output_types = ['prediction']
        else:
            loss = SparseSoftmaxCrossEntropy()
            output_types = ['prediction', 'loss']
        super(GCNModel, self).__init__(model,
                                       loss=loss,
                                       output_types=output_types,
                                       **kwargs)

        self._self_loop = self_loop

    def _prepare_batch(self, batch):
        """Create batch data for GCN.

        Parameters
        ----------
        batch: tuple
            The tuple is ``(inputs, labels, weights)``.

        Returns
        -------
        inputs: DGLGraph
            DGLGraph for a batch of graphs.
        labels: list of torch.Tensor or None
            The graph labels.
        weights: list of torch.Tensor or None
            The weights for each sample or sample/task pair converted to torch.Tensor.
        """
        try:
            import dgl
        except:
            raise ImportError('This class requires dgl.')

        inputs, labels, weights = batch
        dgl_graphs = [
            graph.to_dgl_graph(self_loop=self._self_loop) for graph in inputs[0]
        ]
        inputs = dgl.batch(dgl_graphs).to(self.device)
        _, labels, weights = super(GCNModel, self)._prepare_batch(
            ([], labels, weights))
        return inputs, labels, weights



"""
Compute various spatial fingerprints for macromolecular complexes.
"""
import itertools
import logging
import numpy as np
from deepchem.utils import rdkit_utils
from deepchem.feat import ComplexFeaturizer
from deepchem.utils.voxel_utils import voxelize
from deepchem.utils.voxel_utils import convert_atom_to_voxel
from deepchem.utils.voxel_utils import convert_atom_pair_to_voxel
from deepchem.utils.noncovalent_utils import compute_salt_bridges
from deepchem.utils.noncovalent_utils import compute_binding_pocket_cation_pi
from deepchem.utils.noncovalent_utils import compute_pi_stack
from deepchem.utils.noncovalent_utils import compute_hydrogen_bonds
from deepchem.utils.rdkit_utils import MoleculeLoadException
from deepchem.utils.rdkit_utils import compute_contact_centroid
from deepchem.utils.geometry_utils import compute_pairwise_distances
from deepchem.utils.geometry_utils import subtract_centroid
from deepchem.utils.fragment_utils import get_partial_charge
from deepchem.utils.fragment_utils import reduce_molecular_complex_to_contacts
from typing import List, Tuple, Optional

logger = logging.getLogger(__name__)

HBOND_DIST_BINS = [(2.2, 2.5), (2.5, 3.2), (3.2, 4.0)]
HBOND_ANGLE_CUTOFFS = [5., 50., 90.]


def compute_charge_dictionary(molecule):
    """Create a dictionary with partial charges for each atom in the molecule.

    This function assumes that the charges for the molecule are
    already computed (it can be done with
    rdkit_util.compute_charges(molecule))
    """

    charge_dictionary = {}
    for i, atom in enumerate(molecule.GetAtoms()):
        charge_dictionary[i] = get_partial_charge(atom)
    return charge_dictionary


class ChargeVoxelizer(ComplexFeaturizer):
    """Localize partial charges of atoms in macromolecular complexes.

    Given a macromolecular complex made up of multiple
    constitutent molecules, compute the partial (Gasteiger
    charge) on each molecule. For each atom, localize this
    partial charge in the voxel in which it originated to create
    a local charge array. Sum contributions to get an effective
    charge at each voxel.

    Let `voxels_per_edge = int(box_width/voxel_width)`.  Creates a
    tensor output of shape `(voxels_per_edge, voxels_per_edge,
    voxels_per_edge, 1)` for each macromolecular complex that computes
    the effective charge at each voxel.
    """

    def __init__(self,
                 cutoff: float = 4.5,
                 box_width: float = 16.0,
                 voxel_width: float = 1.0,
                 reduce_to_contacts: bool = True):
        """
        Parameters
        ----------
        cutoff: float (default 4.5)
            Distance cutoff in angstroms for molecules in complex.
        box_width: float, optional (default 16.0)
            Size of a box in which voxel features are calculated. Box
            is centered on a ligand centroid.
        voxel_width: float, optional (default 1.0)
            Size of a 3D voxel in a grid.
        reduce_to_contacts: bool, optional
            If True, reduce the atoms in the complex to those near a contact
            region.
        """
        self.cutoff = cutoff
        self.box_width = box_width
        self.voxel_width = voxel_width
        self.reduce_to_contacts = reduce_to_contacts

    def _featurize(self, datapoint, **kwargs):  # -> Optional[np.ndarray]:
        """
        Compute featurization for a single mol/protein complex

        Parameters
        ----------
        datapoint: Tuple[str, str]
            Filenames for molecule and protein.
        """
        if 'complex' in kwargs:
            datapoint = kwargs.get("complex")
            raise DeprecationWarning(
                'Complex is being phased out as a parameter, please pass "datapoint" instead.'
            )

        try:
            fragments = rdkit_utils.load_complex(datapoint, add_hydrogens=False)

        except MoleculeLoadException:
            logger.warning(
                "This molecule cannot be loaded by Rdkit. Returning None")
            return None
        pairwise_features = []
        # We compute pairwise contact fingerprints
        centroid = compute_contact_centroid(fragments, cutoff=self.cutoff)
        if self.reduce_to_contacts:
            fragments = reduce_molecular_complex_to_contacts(
                fragments, self.cutoff)
        # We compute pairwise contact fingerprints
        for (frag1_ind,
             frag2_ind) in itertools.combinations(range(len(fragments)), 2):
            frag1, frag2 = fragments[frag1_ind], fragments[frag2_ind]
            frag1_xyz = subtract_centroid(frag1[0], centroid)
            frag2_xyz = subtract_centroid(frag2[0], centroid)
            xyzs = [frag1_xyz, frag2_xyz]
            rdks = [frag1[1], frag2[1]]
            pairwise_features.append(
                sum([
                    voxelize(convert_atom_to_voxel,
                             hash_function=None,
                             coordinates=xyz,
                             box_width=self.box_width,
                             voxel_width=self.voxel_width,
                             feature_dict=compute_charge_dictionary(mol),
                             nb_channel=1,
                             dtype="np.float16")
                    for xyz, mol in zip(xyzs, rdks)
                ]))
        # Features are of shape (voxels_per_edge, voxels_per_edge, voxels_per_edge, 1) so we should concatenate on the last axis.
        return np.concatenate(pairwise_features, axis=-1)


class SaltBridgeVoxelizer(ComplexFeaturizer):
    """Localize salt bridges between atoms in macromolecular complexes.

    Given a macromolecular complex made up of multiple
    constitutent molecules, compute salt bridges between atoms in
    the macromolecular complex. For each atom, localize this salt
    bridge in the voxel in which it originated to create a local
    salt bridge array. Note that if atoms in two different voxels
    interact in a salt-bridge, the interaction is double counted
    in both voxels.

    Let `voxels_per_edge = int(box_width/voxel_width)`.  Creates a
    tensor output of shape `(voxels_per_edge, voxels_per_edge,
    voxels_per_edge, 1)` for each macromolecular the number of salt
    bridges at each voxel.
    """

    def __init__(self,
                 cutoff: float = 5.0,
                 box_width: float = 16.0,
                 voxel_width: float = 1.0,
                 reduce_to_contacts: bool = True):
        """
        Parameters
        ----------
        cutoff: float, optional (default 5.0)
            The distance in angstroms within which atoms must be to
            be considered for a salt bridge between them.
        box_width: float, optional (default 16.0)
            Size of a box in which voxel features are calculated. Box
            is centered on a ligand centroid.
        voxel_width: float, optional (default 1.0)
            Size of a 3D voxel in a grid.
        reduce_to_contacts: bool, optional
            If True, reduce the atoms in the complex to those near a contact
            region.
        """
        self.cutoff = cutoff
        self.box_width = box_width
        self.voxel_width = voxel_width
        self.reduce_to_contacts = reduce_to_contacts

    def _featurize(self, datapoint, **kwargs):  # -> Optional[np.ndarray]:
        """
        Compute featurization for a single mol/protein complex

        Parameters
        ----------
        datapoint: Tuple[str, str]
            Filenames for molecule and protein.
        """
        if 'complex' in kwargs:
            datapoint = kwargs.get("complex")
            raise DeprecationWarning(
                'Complex is being phased out as a parameter, please pass "datapoint" instead.'
            )

        try:
            fragments = rdkit_utils.load_complex(datapoint, add_hydrogens=False)

        except MoleculeLoadException:
            logger.warning(
                "This molecule cannot be loaded by Rdkit. Returning None")
            return None
        pairwise_features = []
        # We compute pairwise contact fingerprints
        centroid = compute_contact_centroid(fragments, cutoff=self.cutoff)
        if self.reduce_to_contacts:
            fragments = reduce_molecular_complex_to_contacts(
                fragments, self.cutoff)
        for (frag1_ind,
             frag2_ind) in itertools.combinations(range(len(fragments)), 2):
            frag1, frag2 = fragments[frag1_ind], fragments[frag2_ind]
            distances = compute_pairwise_distances(frag1[0], frag2[0])
            frag1_xyz = subtract_centroid(frag1[0], centroid)
            frag2_xyz = subtract_centroid(frag2[0], centroid)
            xyzs = [frag1_xyz, frag2_xyz]
            # rdks = [frag1[1], frag2[1]]
            pairwise_features.append(
                sum([
                    voxelize(convert_atom_pair_to_voxel,
                             hash_function=None,
                             coordinates=xyz,
                             box_width=self.box_width,
                             voxel_width=self.voxel_width,
                             feature_list=compute_salt_bridges(
                                 frag1[1],
                                 frag2[1],
                                 distances,
                                 cutoff=self.cutoff),
                             nb_channel=1) for xyz in xyzs
                ]))
        # Features are of shape (voxels_per_edge, voxels_per_edge, voxels_per_edge, 1) so we should concatenate on the last axis.
        return np.concatenate(pairwise_features, axis=-1)


class CationPiVoxelizer(ComplexFeaturizer):
    """Localize cation-Pi interactions between atoms in macromolecular complexes.

    Given a macromolecular complex made up of multiple
    constitutent molecules, compute cation-Pi between atoms in
    the macromolecular complex. For each atom, localize this salt
    bridge in the voxel in which it originated to create a local
    cation-pi array.

    Let `voxels_per_edge = int(box_width/voxel_width)`.  Creates a
    tensor output of shape `(voxels_per_edge, voxels_per_edge,
    voxels_per_edge, 1)` for each macromolecular complex that counts the
    number of cation-pi interactions at each voxel.
    """

    def __init__(self,
                 cutoff: float = 6.5,
                 angle_cutoff: float = 30.0,
                 box_width: float = 16.0,
                 voxel_width: float = 1.0):
        """
        Parameters
        ----------
        cutoff: float, optional (default 6.5)
            The distance in angstroms within which atoms must be to
            be considered for a cation-pi interaction between them.
        angle_cutoff: float, optional (default 30.0)
            Angle cutoff. Max allowed deviation from the ideal (0deg)
            angle between ring normal and vector pointing from ring
            center to cation (in degrees).
        box_width: float, optional (default 16.0)
            Size of a box in which voxel features are calculated. Box
            is centered on a ligand centroid.
        voxel_width: float, optional (default 1.0)
            Size of a 3D voxel in a grid.
        """
        self.cutoff = cutoff
        self.angle_cutoff = angle_cutoff
        self.box_width = box_width
        self.voxel_width = voxel_width

    def _featurize(self, datapoint, **kwargs):  # -> Optional[np.ndarray]:
        """
        Compute featurization for a single mol/protein complex

        Parameters
        ----------
        datapoint: Tuple[str, str]
            Filenames for molecule and protein.
        """
        if 'complex' in kwargs:
            datapoint = kwargs.get("complex")
            raise DeprecationWarning(
                'Complex is being phased out as a parameter, please pass "datapoint" instead.'
            )

        try:
            fragments = rdkit_utils.load_complex(datapoint, add_hydrogens=False)

        except MoleculeLoadException:
            logger.warning(
                "This molecule cannot be loaded by Rdkit. Returning None")
            return None
        pairwise_features = []
        # We compute pairwise contact fingerprints
        centroid = compute_contact_centroid(fragments, cutoff=self.cutoff)
        for (frag1_ind,
             frag2_ind) in itertools.combinations(range(len(fragments)), 2):
            frag1, frag2 = fragments[frag1_ind], fragments[frag2_ind]
            # distances = compute_pairwise_distances(frag1[0], frag2[0])
            frag1_xyz = subtract_centroid(frag1[0], centroid)
            frag2_xyz = subtract_centroid(frag2[0], centroid)
            xyzs = [frag1_xyz, frag2_xyz]
            # rdks = [frag1[1], frag2[1]]
            pairwise_features.append(
                sum([
                    voxelize(convert_atom_to_voxel,
                             hash_function=None,
                             box_width=self.box_width,
                             voxel_width=self.voxel_width,
                             coordinates=xyz,
                             feature_dict=cation_pi_dict,
                             nb_channel=1) for xyz, cation_pi_dict in zip(
                                 xyzs,
                                 compute_binding_pocket_cation_pi(
                                     frag1[1],
                                     frag2[1],
                                     dist_cutoff=self.cutoff,
                                     angle_cutoff=self.angle_cutoff,
                                 ))
                ]))
        # Features are of shape (voxels_per_edge, voxels_per_edge, voxels_per_edge, 1) so we should concatenate on the last axis.
        return np.concatenate(pairwise_features, axis=-1)


class PiStackVoxelizer(ComplexFeaturizer):
    """Localize Pi stacking interactions between atoms in macromolecular complexes.

    Given a macromolecular complex made up of multiple
    constitutent molecules, compute pi-stacking interactions
    between atoms in the macromolecular complex. For each atom,
    localize this salt bridge in the voxel in which it originated
    to create a local pi-stacking array.

    Let `voxels_per_edge = int(box_width/voxel_width)`.  Creates a
    tensor output of shape `(voxels_per_edge, voxels_per_edge,
    voxels_per_edge, 2)` for each macromolecular complex. Each voxel has
    2 fields, with the first tracking the number of pi-pi parallel
    interactions, and the second tracking the number of pi-T
    interactions.
    """

    def __init__(self,
                 cutoff: float = 4.4,
                 angle_cutoff: float = 30.0,
                 box_width: float = 16.0,
                 voxel_width: float = 1.0):
        """
        Parameters
        ----------
        cutoff: float, optional (default 4.4)
            The distance in angstroms within which atoms must be to
            be considered for a cation-pi interaction between them.
        angle_cutoff: float, optional (default 30.0)
            Angle cutoff. Max allowed deviation from the ideal (0 deg)
            angle between ring normal and vector pointing from ring
            center to other ring center (in degrees).
        box_width: float, optional (default 16.0)
            Size of a box in which voxel features are calculated. Box
            is centered on a ligand centroid.
        voxel_width: float, optional (default 1.0)
            Size of a 3D voxel in a grid.
        """
        self.cutoff = cutoff
        self.angle_cutoff = angle_cutoff
        self.box_width = box_width
        self.voxel_width = voxel_width

    def _featurize(self, datapoint, **kwargs):  # -> Optional[np.ndarray]:
        """
        Compute featurization for a single mol/protein complex

        Parameters
        ----------
        datapoint: Tuple[str, str]
            Filenames for molecule and protein.
        """
        if 'complex' in kwargs:
            datapoint = kwargs.get("complex")
            raise DeprecationWarning(
                'Complex is being phased out as a parameter, please pass "datapoint" instead.'
            )

        try:
            fragments = rdkit_utils.load_complex(datapoint, add_hydrogens=False)

        except MoleculeLoadException:
            logger.warning(
                "This molecule cannot be loaded by Rdkit. Returning None")
            return None
        pairwise_features = []
        # We compute pairwise contact fingerprints
        centroid = compute_contact_centroid(fragments, cutoff=self.cutoff)
        for (frag1_ind,
             frag2_ind) in itertools.combinations(range(len(fragments)), 2):
            frag1, frag2 = fragments[frag1_ind], fragments[frag2_ind]
            distances = compute_pairwise_distances(frag1[0], frag2[0])
            frag1_xyz = subtract_centroid(frag1[0], centroid)
            frag2_xyz = subtract_centroid(frag2[0], centroid)
            xyzs = [frag1_xyz, frag2_xyz]
            # rdks = [frag1[1], frag2[1]]
            protein_pi_t, protein_pi_parallel, ligand_pi_t, ligand_pi_parallel = (
                compute_pi_stack(frag1[1],
                                 frag2[1],
                                 distances,
                                 dist_cutoff=self.cutoff,
                                 angle_cutoff=self.angle_cutoff))
            pi_parallel_tensor = sum([
                voxelize(convert_atom_to_voxel,
                         hash_function=None,
                         box_width=self.box_width,
                         voxel_width=self.voxel_width,
                         coordinates=xyz,
                         feature_dict=feature_dict,
                         nb_channel=1)
                for (xyz, feature_dict
                    ) in zip(xyzs, [ligand_pi_parallel, protein_pi_parallel])
            ])

            pi_t_tensor = sum([
                voxelize(convert_atom_to_voxel,
                         hash_function=None,
                         box_width=self.box_width,
                         voxel_width=self.voxel_width,
                         coordinates=frag1_xyz,
                         feature_dict=protein_pi_t,
                         nb_channel=1)
                for (xyz,
                     feature_dict) in zip(xyzs, [ligand_pi_t, protein_pi_t])
            ])

            pairwise_features.append(
                np.concatenate([pi_parallel_tensor, pi_t_tensor], axis=-1))
        # Features are of shape (voxels_per_edge, voxels_per_edge, voxels_per_edge, 2) so we should concatenate on the last axis.
        return np.concatenate(pairwise_features, axis=-1)


class HydrogenBondCounter(ComplexFeaturizer):
    """Counts hydrogen bonds between atoms in macromolecular complexes.

    Given a macromolecular complex made up of multiple
    constitutent molecules, count the number of hydrogen bonds
    between atoms in the macromolecular complex.

    Creates a scalar output of shape `(3,)` (assuming the default value
    ofor `distance_bins` with 3 bins) for each macromolecular complex
    that computes the total number of hydrogen bonds.
    """

    def __init__(
        self,
        cutoff: float = 4.5,
        reduce_to_contacts: bool = True,
        distance_bins: Optional[List[Tuple[float, float]]] = None,
        angle_cutoffs: Optional[List[float]] = None,
    ):
        """
        Parameters
        ----------
        cutoff: float (default 4.5)
            Distance cutoff in angstroms for molecules in complex.
        reduce_to_contacts: bool, optional
            If True, reduce the atoms in the complex to those near a contact
            region.
        distance_bins: list[tuple]
            List of hydgrogen bond distance bins. If not specified is
            set to default
            `[(2.2, 2.5), (2.5, 3.2), (3.2, 4.0)]`.
        angle_cutoffs: list[float]
            List of hydrogen bond angle cutoffs. Max allowed
            deviation from the ideal (180 deg) angle between
            hydrogen-atom1, hydrogen-atom2 vectors.If not specified
            is set to default `[5, 50, 90]`
        """
        self.cutoff = cutoff
        if distance_bins is None:
            self.distance_bins = HBOND_DIST_BINS
        else:
            self.distance_bins = distance_bins
        if angle_cutoffs is None:
            self.angle_cutoffs = HBOND_ANGLE_CUTOFFS
        else:
            self.angle_cutoffs = angle_cutoffs
        self.reduce_to_contacts = reduce_to_contacts

    def _featurize(self, datapoint, **kwargs):  # -> Optional[np.ndarray]:
        """
        Compute featurization for a single mol/protein complex

        Parameters
        ----------
        datapoint: Tuple[str, str]
            Filenames for molecule and protein.
        """
        if 'complex' in kwargs:
            datapoint = kwargs.get("complex")
            raise DeprecationWarning(
                'Complex is being phased out as a parameter, please pass "datapoint" instead.'
            )
        try:
            fragments = rdkit_utils.load_complex(datapoint, add_hydrogens=False)

        except MoleculeLoadException:
            logger.warning(
                "This molecule cannot be loaded by Rdkit. Returning None")
            return None
        pairwise_features = []
        # We compute pairwise contact fingerprints
        # centroid = compute_contact_centroid(fragments, cutoff=self.cutoff)
        if self.reduce_to_contacts:
            fragments = reduce_molecular_complex_to_contacts(
                fragments, self.cutoff)
        # We compute pairwise contact fingerprints
        for (frag1_ind,
             frag2_ind) in itertools.combinations(range(len(fragments)), 2):
            frag1, frag2 = fragments[frag1_ind], fragments[frag2_ind]
            distances = compute_pairwise_distances(frag1[0], frag2[0])
            # frag1_xyz = subtract_centroid(frag1[0], centroid)
            # frag2_xyz = subtract_centroid(frag2[0], centroid)
            # xyzs = [frag1_xyz, frag2_xyz]
            # rdks = [frag1[1], frag2[1]]
            pairwise_features.append(
                np.concatenate([
                    np.array([len(hbond_list)])
                    for hbond_list in compute_hydrogen_bonds(
                        frag1, frag2, distances, self.distance_bins,
                        self.angle_cutoffs)
                ],
                               axis=-1))
        # Features are of shape (voxels_per_edge, voxels_per_edge, voxels_per_edge, 1) so we should concatenate on the last axis.
        return np.concatenate(pairwise_features, axis=-1)


class HydrogenBondVoxelizer(ComplexFeaturizer):
    """Localize hydrogen bonds between atoms in macromolecular complexes.

    Given a macromolecular complex made up of multiple
    constitutent molecules, compute hydrogen bonds between atoms
    in the macromolecular complex. For each atom, localize this
    hydrogen bond in the voxel in which it originated to create a
    local hydrogen bond array. Note that if atoms in two
    different voxels interact in a hydrogen bond, the interaction
    is double counted in both voxels.

    Let `voxels_per_edge = int(box_width/voxel_width)`.  Creates a
    tensor output of shape `(voxels_per_edge, voxels_per_edge,
    voxels_per_edge, 3)` (assuming the default for `distance_bins` which
    has 3 bins) for each macromolecular complex that counts the number
    of hydrogen bonds at each voxel.
    """

    def __init__(
        self,
        cutoff: float = 4.5,
        box_width: float = 16.0,
        voxel_width: float = 1.0,
        reduce_to_contacts: bool = True,
        distance_bins: Optional[List[Tuple[float, float]]] = None,
        angle_cutoffs: Optional[List[float]] = None,
    ):
        """
        Parameters
        ----------
        cutoff: float (default 4.5)
            Distance cutoff in angstroms for contact atoms in complex.
        box_width: float, optional (default 16.0)
            Size of a box in which voxel features are calculated. Box
            is centered on a ligand centroid.
        voxel_width: float, optional (default 1.0)
            Size of a 3D voxel in a grid.
        reduce_to_contacts: bool, optional
            If True, reduce the atoms in the complex to those near a contact
            region.
        distance_bins: list[tuple]
            List of hydgrogen bond distance bins. If not specified is
            set to default
            `[(2.2, 2.5), (2.5, 3.2), (3.2, 4.0)]`.
        angle_cutoffs: list[float]
            List of hydrogen bond angle cutoffs. Max allowed
            deviation from the ideal (180 deg) angle between
            hydrogen-atom1, hydrogen-atom2 vectors.If not specified
            is set to default `[5, 50, 90]`
        """
        self.cutoff = cutoff
        if distance_bins is None:
            self.distance_bins = HBOND_DIST_BINS
        else:
            self.distance_bins = distance_bins
        if angle_cutoffs is None:
            self.angle_cutoffs = HBOND_ANGLE_CUTOFFS
        else:
            self.angle_cutoffs = angle_cutoffs
        self.box_width = box_width
        self.voxel_width = voxel_width
        self.reduce_to_contacts = reduce_to_contacts

    def _featurize(self, datapoint, **kwargs):  # -> Optional[np.ndarray]:
        """
        Compute featurization for a single mol/protein complex

        Parameters
        ----------
        datapoint: Tuple[str, str]
            Filenames for molecule and protein.
        """
        if 'complex' in kwargs:
            datapoint = kwargs.get("complex")
            raise DeprecationWarning(
                'Complex is being phased out as a parameter, please pass "datapoint" instead.'
            )
        try:
            fragments = rdkit_utils.load_complex(datapoint, add_hydrogens=False)

        except MoleculeLoadException:
            logger.warning(
                "This molecule cannot be loaded by Rdkit. Returning None")
            return None
        pairwise_features = []
        # We compute pairwise contact fingerprints
        centroid = compute_contact_centroid(fragments, cutoff=self.cutoff)
        if self.reduce_to_contacts:
            fragments = reduce_molecular_complex_to_contacts(
                fragments, self.cutoff)
        for (frag1_ind,
             frag2_ind) in itertools.combinations(range(len(fragments)), 2):
            frag1, frag2 = fragments[frag1_ind], fragments[frag2_ind]
            distances = compute_pairwise_distances(frag1[0], frag2[0])
            frag1_xyz = subtract_centroid(frag1[0], centroid)
            frag2_xyz = subtract_centroid(frag2[0], centroid)
            xyzs = [frag1_xyz, frag2_xyz]
            # rdks = [frag1[1], frag2[1]]
            pairwise_features.append(
                np.concatenate([
                    sum([
                        voxelize(convert_atom_pair_to_voxel,
                                 hash_function=None,
                                 box_width=self.box_width,
                                 voxel_width=self.voxel_width,
                                 coordinates=xyz,
                                 feature_list=hbond_list,
                                 nb_channel=1) for xyz in xyzs
                    ]) for hbond_list in compute_hydrogen_bonds(
                        frag1, frag2, distances, self.distance_bins,
                        self.angle_cutoffs)
                ],
                               axis=-1))
        # Features are of shape (voxels_per_edge, voxels_per_edge, voxels_per_edge, 1) so we should concatenate on the last axis.
        return np.concatenate(pairwise_features, axis=-1)



import numpy as np
import tensorflow as tf
from collections.abc import Sequence as SequenceCollection

import logging
from deepchem.models import KerasModel, layers
from deepchem.models.losses import L2Loss, SparseSoftmaxCrossEntropy
from deepchem.models.keras_model import _StandardLoss
from tensorflow.keras.layers import Input, Dense, Dropout, ReLU, Concatenate, Add, Multiply, Softmax

logger = logging.getLogger(__name__)


class ProgressiveMultitaskRegressor(KerasModel):
    """Implements a progressive multitask neural network for regression.

    Progressive networks allow for multitask learning where each task
    gets a new column of weights. As a result, there is no exponential
    forgetting where previous tasks are ignored.

    References
    ----------
    See [1]_ for a full description of the progressive architecture

    .. [1] Rusu, Andrei A., et al. "Progressive neural networks." arXiv preprint
        arXiv:1606.04671 (2016).
    """

    def __init__(self,
                 n_tasks,
                 n_features,
                 alpha_init_stddevs=0.02,
                 layer_sizes=[1000],
                 weight_init_stddevs=0.02,
                 bias_init_consts=1.0,
                 weight_decay_penalty=0.0,
                 weight_decay_penalty_type="l2",
                 dropouts=0.5,
                 activation_fns=tf.nn.relu,
                 n_outputs=1,
                 **kwargs):
        """Creates a progressive network.

        Only listing parameters specific to progressive networks here.

        Parameters
        ----------
        n_tasks: int
            Number of tasks
        n_features: int
            Number of input features
        alpha_init_stddevs: list
            List of standard-deviations for alpha in adapter layers.
        layer_sizes: list
            the size of each dense layer in the network.  The length of this list determines the number of layers.
        weight_init_stddevs: list or float
            the standard deviation of the distribution to use for weight initialization of each layer.  The length
            of this list should equal len(layer_sizes)+1.  The final element corresponds to the output layer.
            Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
        bias_init_consts: list or float
            the value to initialize the biases in each layer to.  The length of this list should equal len(layer_sizes)+1.
            The final element corresponds to the output layer.  Alternatively this may be a single value instead of a list,
            in which case the same value is used for every layer.
        weight_decay_penalty: float
            the magnitude of the weight decay penalty to use
        weight_decay_penalty_type: str
            the type of penalty to use for weight decay, either 'l1' or 'l2'
        dropouts: list or float
            the dropout probablity to use for each layer.  The length of this list should equal len(layer_sizes).
            Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
        activation_fns: list or object
            the Tensorflow activation function to apply to each layer.  The length of this list should equal
            len(layer_sizes).  Alternatively this may be a single value instead of a list, in which case the
            same value is used for every layer.
        """

        if weight_decay_penalty != 0.0:
            raise ValueError('Weight decay is not currently supported')
        self.n_tasks = n_tasks
        self.n_features = n_features
        self.layer_sizes = layer_sizes
        self.alpha_init_stddevs = alpha_init_stddevs
        self.weight_init_stddevs = weight_init_stddevs
        self.bias_init_consts = bias_init_consts
        self.dropouts = dropouts
        self.activation_fns = activation_fns
        self.n_outputs = n_outputs

        n_layers = len(layer_sizes)
        if not isinstance(weight_init_stddevs, SequenceCollection):
            self.weight_init_stddevs = [weight_init_stddevs] * n_layers
        if not isinstance(alpha_init_stddevs, SequenceCollection):
            self.alpha_init_stddevs = [alpha_init_stddevs] * n_layers
        if not isinstance(bias_init_consts, SequenceCollection):
            self.bias_init_consts = [bias_init_consts] * n_layers
        if not isinstance(dropouts, SequenceCollection):
            self.dropouts = [dropouts] * n_layers
        if not isinstance(activation_fns, SequenceCollection):
            self.activation_fns = [activation_fns] * n_layers

        # Add the input features.
        mol_features = Input(shape=(n_features,))

        all_layers = {}
        outputs = []
        self._task_layers = []
        for task in range(self.n_tasks):
            task_layers = []
            for i in range(n_layers):
                if i == 0:
                    prev_layer = mol_features
                else:
                    prev_layer = all_layers[(i - 1, task)]
                    if task > 0:
                        lateral_contrib, trainables = self.add_adapter(
                            all_layers, task, i)
                        task_layers.extend(trainables)

                dense = Dense(
                    layer_sizes[i],
                    kernel_initializer=tf.keras.initializers.TruncatedNormal(
                        stddev=self.weight_init_stddevs[i]),
                    bias_initializer=tf.constant_initializer(
                        value=self.bias_init_consts[i]))
                layer = dense(prev_layer)
                task_layers.append(dense)

                if i > 0 and task > 0:
                    layer = Add()([layer, lateral_contrib])
                assert self.activation_fns[
                    i] is tf.nn.relu, "Only ReLU is supported"
                layer = ReLU()(layer)
                if self.dropouts[i] > 0.0:
                    layer = Dropout(self.dropouts[i])(layer)
                all_layers[(i, task)] = layer

            prev_layer = all_layers[(n_layers - 1, task)]
            dense = Dense(
                n_outputs,
                kernel_initializer=tf.keras.initializers.TruncatedNormal(
                    stddev=self.weight_init_stddevs[-1]),
                bias_initializer=tf.constant_initializer(
                    value=self.bias_init_consts[-1]))
            layer = dense(prev_layer)
            task_layers.append(dense)

            if task > 0:
                lateral_contrib, trainables = self.add_adapter(
                    all_layers, task, n_layers)
                task_layers.extend(trainables)
                layer = Add()([layer, lateral_contrib])
            output_layer = self.create_output(layer)
            outputs.append(output_layer)
            self._task_layers.append(task_layers)

        outputs = layers.Stack(axis=1)(outputs)
        model = tf.keras.Model(inputs=mol_features, outputs=outputs)
        super(ProgressiveMultitaskRegressor,
              self).__init__(model, self.create_loss(), **kwargs)

    def create_loss(self):
        return L2Loss()

    def create_output(self, layer):
        return layer

    def add_adapter(self, all_layers, task, layer_num):
        """Add an adapter connection for given task/layer combo"""
        i = layer_num
        prev_layers = []
        trainable_layers = []
        # Handle output layer
        if i < len(self.layer_sizes):
            layer_sizes = self.layer_sizes
            alpha_init_stddev = self.alpha_init_stddevs[i]
            weight_init_stddev = self.weight_init_stddevs[i]
            bias_init_const = self.bias_init_consts[i]
        elif i == len(self.layer_sizes):
            layer_sizes = self.layer_sizes + [self.n_outputs]
            alpha_init_stddev = self.alpha_init_stddevs[-1]
            weight_init_stddev = self.weight_init_stddevs[-1]
            bias_init_const = self.bias_init_consts[-1]
        else:
            raise ValueError("layer_num too large for add_adapter.")
        # Iterate over all previous tasks.
        for prev_task in range(task):
            prev_layers.append(all_layers[(i - 1, prev_task)])
        # prev_layers is a list with elements of size
        # (batch_size, layer_sizes[i-1])
        if len(prev_layers) == 1:
            prev_layer = prev_layers[0]
        else:
            prev_layer = Concatenate(axis=1)(prev_layers)
        alpha = layers.Variable(
            tf.random.truncated_normal((1,), stddev=alpha_init_stddev))
        trainable_layers.append(alpha)

        prev_layer = Multiply()([prev_layer, alpha([prev_layer])])
        dense1 = Dense(
            layer_sizes[i - 1],
            kernel_initializer=tf.keras.initializers.TruncatedNormal(
                stddev=weight_init_stddev),
            bias_initializer=tf.constant_initializer(value=bias_init_const))
        prev_layer = dense1(prev_layer)
        trainable_layers.append(dense1)

        dense2 = Dense(layer_sizes[i],
                       kernel_initializer=tf.keras.initializers.TruncatedNormal(
                           stddev=weight_init_stddev),
                       use_bias=False)
        prev_layer = dense2(prev_layer)
        trainable_layers.append(dense2)

        return prev_layer, trainable_layers

    def fit(self,
            dataset,
            nb_epoch=10,
            max_checkpoints_to_keep=5,
            checkpoint_interval=1000,
            deterministic=False,
            restore=False,
            **kwargs):
        for task in range(self.n_tasks):
            self.fit_task(dataset,
                          task,
                          nb_epoch=nb_epoch,
                          max_checkpoints_to_keep=max_checkpoints_to_keep,
                          checkpoint_interval=checkpoint_interval,
                          deterministic=deterministic,
                          restore=restore,
                          **kwargs)

    def fit_task(self,
                 dataset,
                 task,
                 nb_epoch=10,
                 max_checkpoints_to_keep=5,
                 checkpoint_interval=1000,
                 deterministic=False,
                 restore=False,
                 **kwargs):
        """Fit one task."""
        shape = dataset.get_shape()
        batch = [[np.zeros((self.batch_size,) + s[1:])] for s in shape]
        self._create_training_ops(batch)
        generator = self.default_generator(dataset,
                                           epochs=nb_epoch,
                                           deterministic=deterministic)
        variables = []
        for layer in self._task_layers[task]:
            variables += layer.trainable_variables
        loss = TaskLoss(self.model, self.create_loss(), task)
        self.fit_generator(generator,
                           max_checkpoints_to_keep,
                           checkpoint_interval,
                           restore,
                           variables=variables,
                           loss=loss)


class ProgressiveMultitaskClassifier(ProgressiveMultitaskRegressor):
    """Implements a progressive multitask neural network for classification.

    Progressive Networks: https://arxiv.org/pdf/1606.04671v3.pdf

    Progressive networks allow for multitask learning where each task
    gets a new column of weights. As a result, there is no exponential
    forgetting where previous tasks are ignored.

    """

    def __init__(self,
                 n_tasks,
                 n_features,
                 alpha_init_stddevs=0.02,
                 layer_sizes=[1000],
                 weight_init_stddevs=0.02,
                 bias_init_consts=1.0,
                 weight_decay_penalty=0.0,
                 weight_decay_penalty_type="l2",
                 dropouts=0.5,
                 activation_fns=tf.nn.relu,
                 **kwargs):
        n_outputs = 2
        super(ProgressiveMultitaskClassifier, self).__init__(
            n_tasks,
            n_features,
            alpha_init_stddevs=alpha_init_stddevs,
            layer_sizes=layer_sizes,
            weight_init_stddevs=weight_init_stddevs,
            bias_init_consts=bias_init_consts,
            weight_decay_penalty=weight_decay_penalty,
            weight_decay_penalty_type=weight_decay_penalty_type,
            dropouts=dropouts,
            activation_fns=activation_fns,
            n_outputs=n_outputs,
            **kwargs)

    def create_loss(self):
        return SparseSoftmaxCrossEntropy()

    def create_output(self, layer):
        return Softmax()(layer)


class TaskLoss(_StandardLoss):

    def __init__(self, model, loss, task):
        super(TaskLoss, self).__init__(model, loss)
        self.task = task

    def __call__(self, outputs, labels, weights):
        outputs = [t[:, self.task] for t in outputs]
        labels = [t[:, self.task] for t in labels]
        weights = [t[:, self.task] for t in weights]
        return super(TaskLoss, self).__call__(outputs, labels, weights)
